
import os
import shutil
import tempfile
import logging
from typing import Optional
import uuid
import urllib
import json

import jetpack.config
import jetpack.lockers
import jetpack.settings
import jetpack.util
import jetpack.util.azcopy

logger = logging.getLogger(__name__)


class UploadLogsError(Exception):
    def __init__(self, message: str, local_archive: str = None, remote_archive: str = None):
        super(UploadLogsError, self).__init__(message)
        self.local_archive: Optional[str] = local_archive
        self.remote_archive: Optional[str] = remote_archive


def upload_logs(path: str = None, upload: bool = True, sign: bool = True, client_id_override: str = None) -> str:
    """ Compresses a log directory PATH and then optionally
    uploads the result to Azure Storage as well as generating a temporary signed URL to access the logs for supportability

    :param str path: Path to directory of logs to archive
    :param bool upload: If true, upload log archive to Azure Storage
    :param bool sign: If True, request a signed URL from CycleCloud to the previously uploaded archive
    :return: The path (local or remote) to the archived log files
    """

    # If no path was specified, we'll default to the Jetpack logs directory
    if path is None:
        path = os.path.join(jetpack.util.jetpack_home(), 'logs')

    if not os.path.isdir(path):
        # TODO: Raise a better exception
        raise UploadLogsError("Logs directory '%s' does not exist" % path)

    # The file will be named CLUSTER-NODE-GUID.zip
    # If the node doesn't have an explicit name (i.e. execute-1) it's possible that the system has failed prior to learning it's real name
    # from CycleCloud. If this is the case, fall back on the template name and then simply to 'unknown'.
    cluster_name: str = jetpack.config.get('ec2.userdata.user_data.pool.name', 'unknown')  # TODO: Move this to settings
    node_name: str = jetpack.config.get('cyclecloud.node.name', jetpack.config.get('cyclecloud.node.template', 'unknown'))

    filename: str = "%s-%s-%s" % (cluster_name, node_name, uuid.uuid4())
    local_archive: str = _create_temporary_archive(path, filename)
    dst_name: str = os.path.basename(local_archive)
    remote_archive: Optional[str] = None

    if upload:
        locker_name: str = jetpack.settings.get('locker')
        dst_locker: jetpack.lockers.Locker = jetpack.lockers.get(locker_name)
        dst_path: str = "issues/%s" % dst_name
        try:
            remote_archive = dst_locker.url(dst_path)
            jetpack.util.azcopy.copy(local_archive, dst_path, dst_locker=dst_locker, client_id_override=client_id_override)
        except Exception:
            logger.exception("Unable to upload logs to %s" % remote_archive)
            raise UploadLogsError("Unable to upload logs to %s" % remote_archive, local_archive=local_archive)

        if sign:
            try:
                remote_archive = _sign_url(dst_path, locker_name)
            except Exception:
                logger.exception("Unable to sign uploaded logs at %s" % remote_archive)
                raise UploadLogsError("Unable to sign uploaded logs at %s" % remote_archive, local_archive=local_archive, remote_archive=remote_archive)

    # If we successfully uploaded the file, go ahead an remove it from the local filesystem
    if upload:
        shutil.rmtree(os.path.dirname(local_archive))

    return remote_archive or local_archive


def _create_temporary_archive(path: str, filename: str) -> str:
    """ Creates a zip file of PATH in a temporary directory with a given FILENAME

    :param str path: The path to a directory to archive
    :param str filename: The name of the file to create in a temp directory
    :return: The path to the resulting archive
    """
    tmpdir: str = tempfile.TemporaryDirectory().name
    archive_name: str = os.path.join(tmpdir, filename)
    archive_path: str = shutil.make_archive(archive_name, 'zip', root_dir=path)
    return archive_path


def _sign_url(path: str, locker_name: str) -> str:
    """ Asks CC to sign a URL that is inside a given LOCKER """
    params = {
        'locker': locker_name,
        'path': path
    }
    url = "/clusterlink/sign_url"
    r = jetpack.util.query_cyclecloud(url, params=params, method="POST")
    if r.status != 200:
        raise Exception("Invalid status returned from clusterlink/sign_url: %s" % r.status)
    try:
        data = json.loads(r.read())
        return data['signed_url']
    except ValueError:
        raise  # Just reraise the error so we log it appropriately
