from datetime import datetime
import json
import logging
import os
import sys
import traceback

import jetpack.util
from jetpack.util import UserError

logger = logging.getLogger(__name__)

STATUS_FILE = os.path.join(jetpack.util.jetpack_home(), 'run', 'status.json')

# NOTE: Python2 doesn't support enums by default so we use a class as a poor man's enum
class JetpackResource:
    NodeHealthChecks = "NodeHealthChecks"
    UserAccess = "UserAccess"
    SoftwareConfiguration = "SoftwareConfiguration"
    Unknown = "Unknown"

class JetpackStatus:
    NoStatus = "Pending"
    Error = "Error"
    Warning = "Warning"
    Ok = "OK"

def _set_if_not_none(d, name, value):
    """ Adds a key/value pair to a dictionary (d) if value is defined """
    if value is not None:
        d[name] = value

def _replace(name, status, update_only=False):
    ''' updates the status of 'name' with status (dict), sends the status back to CC unless update_only is True '''
    existing_status = _read_status()
    existing_status[name] = status
    _write_status(existing_status)
    if not update_only:
        send(existing_status)

def set(name, status=None, description=None, detail=None, recommendation=None, message=None, help_link=None, active=None, start_time=None, end_time=None, version=None, update_only=False):
    '''
    Sets the local node status for a given resource

    '''
    s = {}
    _set_if_not_none(s, 'status', status)
    _set_if_not_none(s, 'description', description)
    _set_if_not_none(s, 'detail', detail)
    _set_if_not_none(s, 'recommendation', recommendation)
    _set_if_not_none(s, 'message', message)
    _set_if_not_none(s, 'help_link', help_link)
    _set_if_not_none(s, 'active', active)
    _set_if_not_none(s, 'start_time', start_time)
    _set_if_not_none(s, 'end_time', end_time)
    _set_if_not_none(s, 'version', version)

    _replace(name, s, update_only=update_only)

def update(name, status=None, description=None, detail=None, recommendation=None, message=None, help_link=None, active=None, start_time=None, end_time=None, version=None, update_only=False):
    '''
    Updates the local node status for a given resrouce

    :param name: The name of the resource whose status is being updated
    :type name: str
    :param status: The status of the resource
    :type status: str
    :param description: A description of the status of the resource
    :type description: str
    :param detail: A detailed description of the status (> 1 line)
    :type detail: str
    :param recommendation: A recommendation for how to fix the warning or error status being reported
    :type recommendation: str
    :param message: A short one line message describing the status
    :type message: str
    :param help_link: A link to a resource which describes the problem and how to fix it.
    :type help_link: str
    :param active: A boolean descripting if this is an active operation
    :type active: bool
    :param start_time: A ISO8601 formatted timestamp of the start of the operation
    :type start_time: str
    :param end_time: A ISO8601 formatted timestamp of the end of the operation
    :type end_time: str
    :param version: If given, this is reported back as the currently valid version of this aspect
    :type version: int
    :param update_only: If True, the status update will not be sent to CycleCloud
    :type update_only: bool

    Example usage:
        update(JetpackResource.UserAccess, JetpackStatus.Error, "Error adding user 'test-user', the name is already in use!")
    '''
    existing_status = _read_status()

    # Get the current status for 'name' from the existing_statuses
    s = existing_status.get(name, {})

    # Update them leaving anything already set 'as is'
    _set_if_not_none(s, 'status', status)
    _set_if_not_none(s, 'description', description)
    _set_if_not_none(s, 'detail', detail)
    _set_if_not_none(s, 'recommendation', recommendation)
    _set_if_not_none(s, 'message', message)
    _set_if_not_none(s, 'help_link', help_link)
    _set_if_not_none(s, 'active', active)
    _set_if_not_none(s, 'start_time', start_time)
    _set_if_not_none(s, 'end_time', end_time)
    _set_if_not_none(s, 'version', version)

    _replace(name, s, update_only=update_only)


def update_from_exception(name, e):
    if isinstance(e, UserError):
        update(name, JetpackStatus.Error, e.description, e.detail, e.recommendation, str(e), e.help_link)
    else:
        description = str(e)

        # Attempt to extract a traceback if one is available and put it in the detail
        detail = None
        ex_type, ex, tb = sys.exc_info()
        if tb:
            detail = "".join(traceback.format_tb(tb))

        update(name, JetpackStatus.Error, description, detail, message="Unexpected error")


def send(status=None):
    """ Sends the local node status to CycleCloud, reads status from disk if not passed in """
    if status is None:
        status = _read_status()

    logger.debug("Sending node status to CycleCloud: %s" % json.dumps(status, indent=4, sort_keys=True))
    jetpack.messaging.send_internal_message(status, "status")

def _read_status():
    """ Reads the status from disk and returns the parsed JSON """
    status = {}
    if os.path.isfile(STATUS_FILE):
        with open(STATUS_FILE) as f:
            status = json.loads(f.read())
    return status

def _write_status(status):
    """ Writes the status back to disk """
    if not os.path.isdir(os.path.dirname(STATUS_FILE)):
        os.makedirs(os.path.dirname(STATUS_FILE))

    with open(STATUS_FILE, 'w') as f:
        f.write(json.dumps(status, indent=4, sort_keys=True))
